#include <cstdlib>
#include <fstream>
#include <iostream>
#include <vector>

#include "rlcsa_builder.h"
#include "rlcsa.h"
#include "misc/utils.h"


using namespace CSA;


int
incrementalRLCSA(std::string list_name, std::string base_name, usint block_size, usint sample_rate)
{
  Parameters parameters;
  parameters.set(RLCSA_BLOCK_SIZE.first, block_size);
  parameters.set(SAMPLE_RATE.first, sample_rate);

  if(sample_rate > 0)
  {
    parameters.set(SUPPORT_LOCATE.first, 1);
    parameters.set(SUPPORT_DISPLAY.first, 1);
  }
  else
  {
    parameters.set(SUPPORT_LOCATE.first, 0);
    parameters.set(SUPPORT_DISPLAY.first, 0);
  }

  std::string parameters_name = base_name + PARAMETERS_EXTENSION;
  parameters.print();
  parameters.write(parameters_name);

  std::ifstream filelist(list_name.c_str(), std::ios_base::binary);
  if(!filelist)
  {
    std::cerr << "Error opening file list!" << std::endl;
    return 6;
  }
  std::vector<std::string> files;
  readRows(filelist, files, true);
  filelist.close();
  std::cout << "Preparing to index " << files.size() << " files." << std::endl;
  std::cout << "Index name: " << base_name << std::endl;
  std::cout << std::endl;

  std::clock_t start = std::clock();
  RLCSABuilder builder(parameters.get(RLCSA_BLOCK_SIZE), parameters.get(SAMPLE_RATE), 0);
  usint total_size = 0;

  // Build index for the first file.
  std::vector<std::string>::iterator iter = files.begin();
  std::ifstream first((*iter).c_str(), std::ios_base::binary);
  usint size = fileSize(first);
  std::cout << "Indexing file " << *iter << " (" << size << " bytes)... "; std::cout.flush();
  uchar* data = new uchar[size];
  first.read((char*)data, size);
  first.close();

  builder.insertSequence((char*)data, size, true);
  total_size += size;
  std::cout << "done." << std::endl;
  std::cout << std::endl;

  // Add other files to the index.
  for(iter++; iter != files.end(); iter++)
  {
    // Build index for the increment.
    std::ifstream incr((*iter).c_str(), std::ios_base::binary);
    size = fileSize(incr);
    std::cout << "Indexing file " << *iter << " (" << size << " bytes)... "; std::cout.flush();
    total_size += size;
    data = new uchar[size];
    incr.read((char*)data, size);
    incr.close();
    builder.insertSequence((char*)data, size, true);
    std::cout << "done." << std::endl;
    std::cout << std::endl;
  }

  RLCSA* rlcsa = builder.getRLCSA();
  rlcsa->writeTo(base_name);
  std::clock_t stop = std::clock();
  double build_time = builder.getBuildTime();
  double search_time = builder.getSearchTime();
  double merge_time = builder.getMergeTime();

  double megabytes = total_size / (double)MEGABYTE;
  double time = (stop - start) / (double)CLOCKS_PER_SEC;
  std::cout << megabytes << " megabytes in " << time << " seconds (" << (megabytes / time) << " MB/s)" << std::endl;
  std::cout << "(build " << build_time << " s, search " << search_time << " s, merge " << merge_time << " s)" << std::endl;
  std::cout << std::endl;

  delete rlcsa;
  return 0;
}


int
lineByLineRLCSA(std::string base_name, usint block_size, usint sample_rate, usint buffer_size)
{
  Parameters parameters;
  parameters.set(RLCSA_BLOCK_SIZE.first, block_size);
  parameters.set(SAMPLE_RATE.first, sample_rate);

  if(sample_rate > 0)
  {
    parameters.set(SUPPORT_LOCATE.first, 1);
    parameters.set(SUPPORT_DISPLAY.first, 1);
  }
  else
  {
    parameters.set(SUPPORT_LOCATE.first, 0);
    parameters.set(SUPPORT_DISPLAY.first, 0);
  }

  std::string parameters_name = base_name + PARAMETERS_EXTENSION;
  parameters.print();
  parameters.write(parameters_name);

  std::cout << "Input: " << base_name << std::endl;
  std::ifstream input_file(base_name.c_str(), std::ios_base::binary);
  if(!input_file)
  {
    std::cerr << "Error opening input file!" << std::endl;
    return 7;
  }
  std::cout << "Buffer size: " << buffer_size << " MB" << std::endl;
  std::cout << std::endl;

  std::clock_t start = std::clock();
  RLCSABuilder builder(parameters.get(RLCSA_BLOCK_SIZE), parameters.get(SAMPLE_RATE), buffer_size * MEGABYTE);

  // Build index for the first file.
  usint lines = 0, total = 0;
  while(input_file)
  {
    char buffer[4096];  // FIXME What if lines are longer?
    input_file.getline(buffer, 4096);
    usint chars = input_file.gcount();
    lines++; total += chars;
    if(chars > 1) { builder.insertSequence(buffer, chars - 1, false); }
  }

  RLCSA* rlcsa = builder.getRLCSA();
  rlcsa->writeTo(base_name);
  std::clock_t stop = std::clock();
  double build_time = builder.getBuildTime();
  double search_time = builder.getSearchTime();
  double merge_time = builder.getMergeTime();

  double megabytes = rlcsa->getSize() / (double)MEGABYTE;
  usint sequences = rlcsa->getNumberOfSequences();
  double time = (stop - start) / (double)CLOCKS_PER_SEC;
  std::cout << sequences << " sequences" << std::endl;
  std::cout << megabytes << " megabytes in " << time << " seconds (" << (megabytes / time) << " MB/s)" << std::endl;
  std::cout << "(build " << build_time << " s, search " << search_time << " s, merge " << merge_time << " s)" << std::endl;
  std::cout << std::endl;

  delete rlcsa;
  return 0;
}


int
main(int argc, char** argv)
{
  std::cout << "RLCSA builder" << std::endl;
  if(argc < 2)
  {
    std::cout << "Usage:" << std::endl;
    std::cout << "  Standard:      rlcsa_builder base_name [block_size [sample_rate]]" << std::endl;
    std::cout << "  Incremental:   rlcsa_builder -i list base_name [block_size [sample_rate]]" << std::endl;
    std::cout << "  Line by line:  rlcsa_builder -l base_name buffer_size [block_size [sample_rate]]" << std::endl;
    return 1;
  }

  int name_arg, block_arg, sample_arg;
  bool incremental = ((std::string("-i")).compare(argv[1]) == 0);
  bool line_by_line = ((std::string("-l")).compare(argv[1]) == 0);
  if(incremental)
  {
    name_arg = 3; block_arg = 4; sample_arg = 5;
  }
  else if(line_by_line)
  {
    name_arg = 2; block_arg = 4; sample_arg = 5;
  }
  else
  {
    name_arg = 1; block_arg = 2; sample_arg = 3;
  }

  if(incremental)
  {
    std::cout << "Using incremental construction." << std::endl;
    std::cout << std::endl;
    if(argc < name_arg + 1)
    {
      std::cout << "Usage: rlcsa_builder -i list base_name [block_size [sample_rate]]" << std::endl;
      return 3;
    }

    usint block_size = (argc > block_arg ? atoi(argv[block_arg]) : RLCSA_BLOCK_SIZE.second);
    usint sample_rate = (argc > sample_arg ? atoi(argv[sample_arg]) : SAMPLE_RATE.second);
    return incrementalRLCSA(argv[2], argv[name_arg], block_size, sample_rate);
  }

  if(line_by_line)
  {
    std::cout << "Using line by line construction." << std::endl;
    std::cout << std::endl;
    if(argc < name_arg + 1)
    {
      std::cout << "Usage: rlcsa_builder -l base_name [block_size [sample_rate]]" << std::endl;
      return 4;
    }

    usint block_size = (argc > block_arg ? atoi(argv[block_arg]) : RLCSA_BLOCK_SIZE.second);
    usint sample_rate = (argc > sample_arg ? atoi(argv[sample_arg]) : SAMPLE_RATE.second);
    return lineByLineRLCSA(argv[name_arg], block_size, sample_rate, atoi(argv[3]));
  }

  std::string base_name = argv[name_arg];
  std::cout << "Input: " << base_name << std::endl;
  std::ifstream input_file(base_name.c_str(), std::ios_base::binary);
  if(!input_file)
  {
    std::cerr << "Error opening input file!" << std::endl;
    return 5;
  }
  usint size = fileSize(input_file);

  Parameters parameters;
  parameters.set(RLCSA_BLOCK_SIZE);
  parameters.set(SAMPLE_RATE);
  std::string parameters_name = base_name + PARAMETERS_EXTENSION;
  parameters.read(parameters_name);

  if(argc > block_arg)
  {
    parameters.set(RLCSA_BLOCK_SIZE.first, atoi(argv[block_arg]));
  }
  if(argc > sample_arg)
  {
    parameters.set(SAMPLE_RATE.first, atoi(argv[sample_arg]));
  }
  if(parameters.get(SAMPLE_RATE) > 0)
  {
    parameters.set(SUPPORT_LOCATE.first, 1);
    parameters.set(SUPPORT_DISPLAY.first, 1);
  }
  else
  {
    parameters.set(SUPPORT_LOCATE.first, 0);
    parameters.set(SUPPORT_DISPLAY.first, 0);
  }
  std::cout << std::endl;

  parameters.print();
  parameters.write(parameters_name);

  std::clock_t start = std::clock();

  uchar* data = new uchar[size];
  input_file.read((char*)data, size);
  input_file.close();

  // FIXME First true is support multiple sequences. Should add as a parameter.
  RLCSA rlcsa(data, size, parameters.get(RLCSA_BLOCK_SIZE), parameters.get(SAMPLE_RATE), true, true);
  if(rlcsa.isOk()) { rlcsa.writeTo(argv[1]); }

  std::clock_t stop = std::clock();
  double megabytes = size / (double)MEGABYTE;
  double time = (stop - start) / (double)CLOCKS_PER_SEC;
  std::cout << megabytes << " megabytes in " << time << " seconds (" << (megabytes / time) << " MB/s)" << std::endl;
  std::cout << std::endl;

  return 0;
}
